# A list of functions used to compute the quantile mapping on a 91-day moving window, 
# using .01 to .99 quantiles for deriving the correction 
# applying a frequency adjustment for precipitation
# functions derived from https://github.com/SvenKotlarski/qmCH2018 with small changes

qm.cdf.doy <- function(series, yearbegin, yearend, doywindow, minq, maxq, incq, var, pr.wet){
  
  d		<- qm.doywindowgen(doywindow)
  cdf_doy		<- array(NA,dim=c(365,length(seq(minq,maxq,by=incq))))
  doy.series	<- qm.doystring(yearbegin,yearend)
  fwet_doy	<- rep(NA,365)
  cdfwet_doy	<- array(NA,dim=c(365,101))
  
  # *** set 29th February to 1st March ***
  doy.series[doy.series > 500] <- 60
  
  # *** check of input data: length OK? ***
  if(length(series) != length(doy.series)){
    print("ERROR: data series and doy series not equally long")
    cdf_doy <- NULL
  }else{
    for(i in 1:365){
      #			*** match: returns vector of length(doy.series): Position of matches in d$doy.matrix[i,] or NA if no match
      #			*** => all positions in doy.series that correspond to the window defined in d$doy.matrix[i,] are marked
      cdf_doy[i,] <- quantile(series[which(is.na(match(doy.series,d$doy.matrix[i,])) == FALSE)], probs=seq(minq,maxq,by=incq),na.rm=TRUE)
      
      #			*** possible (shorter) alternative ***
      #			cdf_doy[i,] <- quantile(series[which(!is.na(match(doy.series,d$doy.matrix[i,])))],probs=c(seq(minq,maxq,by=incq)),na.rm=TRUE)
      
      #			*** for precipitation additionally compute wet day frequency and wet day distribution for each DOY ***
      if (var == 'pr') {
        # *** all values for current DOY including window ***
        series.doywin <- series[which(is.na(match(doy.series,d$doy.matrix[i,])) == FALSE)]
        # *** total number of days w/o NA ***
        ndays <- length(series.doywin[!is.na(series.doywin)])
        # *** number of wet days w/o NA ***
        ndays.wet <- length(which(series.doywin >= pr.wet))
        # *** wet day fraction for current DOY ***
        fwet_doy[i] <- ndays.wet / ndays
        # *** all values for current DOY including window but excluding dry days ***
        series.doywin.wet <- series.doywin[series.doywin >= pr.wet]
        # *** wet day CDF for current DOY including window ***
        cdfwet_doy[i,] <- quantile(series.doywin.wet,probs=seq(0,1,0.01),na.rm=TRUE)
      }
    }
  }
  return(list(cdf.matrix = cdf_doy, f.wet = fwet_doy, cdf.wet = cdfwet_doy))
}


qm.empqm.doy.fa <- function(series, doy.series, cdf.mod, cdf.obs, cdf.obs.wet, cor.fun, fwet.obs, fwet.mod, pr.wet, method=c("linear","bin"), var, minq, maxq, incq){

  if(length(series) != length(doy.series)){ # *** a little check
    print(paste("ERROR: length of time series (",length(series),") and DOY series (",length(doy.series),") do not match.", sep=""))
  }else{
    series_q		<- rep(NA,length(series))
    series_c		<- rep(NA,length(series))
    for(i in 1:length(series)){
      doy <- doy.series[i]
      if(doy == 999){ doy <- 60 }
      if(is.na(series[i]) == "TRUE"){
        series_q[i] <- NA
        series_c[i] <- NA
      }else{
        # *** OPTION A: if dry day in model and if modelled wet day frequency <= observed wet day frequency apply frenquency adjustment ***
        if ((series[i]==0) & (fwet.mod[doy] <= fwet.obs[doy])) {
          # *** generate random number ***
          rnum <- runif(1)
          # *** determine fraction of dry days in model that needs to become wet in order to represent observed wet day frequency ***
          probwet <- 1 - ((1-fwet.obs[doy])/(1-fwet.mod[doy]))
          # *** if rnum <= probwet: fill with a precipitation day; else assign original value, i.e. zero precip ***
          if (rnum <= probwet) {
            # *** randomly draw from observed wet day distribution and assign that value ***
            fun <- approxfun(seq(0,1,0.01),cdf.obs.wet[doy,])
            random.number <- runif(1)
            series_c[i] <- fun(random.number)
          }else{
            series_c[i] <- series[i]
          }
          # *** in any case, assign NA to series of modelled quantile numbers: means FA has been carried out! ***
          series_q[i]	<- NA
        }else{
          # *** OPTION B: else, don't apply frenquency adjustment; apply regular correction function ***
          # *** using quantile-"bins" ***
          if(match.arg(method) == "bin"){
            # *** Note: this command implicitly selects the first (last) quantile considered for new extremes , i.e.  ***
            # *** for smaller (larger) values than those occuring in the calbration period (in case of zeros: first   ***
            # *** occurence). If minq=0.01 and maxq=0.99 nex extremes are hence corrected according to the correction ***
            # *** function for the 1st and 99th percentile.                                                           ***
            ix		<- which.min(abs(cdf.mod[doy,]-series[i]))
            # *** Compute quantile index  ***
            series_q[i]	<- minq + (ix-1)*incq
            # *** Compute corrected value ***
            series_c[i]	<- series[i]-cor.fun[doy,ix]
          }
          
          # *** using linear interpolation between quantiles ***
          if(match.arg(method) == "linear"){
            # *** number of interpolated sampling points ***
            npoints <- 1000
            cor		<- approx(cdf.mod[doy,],n=npoints,method="linear")$y
            # *** Note: this command implicitly selects the first (last) quantile considered for new extremes , i.e.  ***
            # *** for smaller (larger) values than those occuring in the calbration period (in case of zeros: first   ***
            # *** occurence). If minq=0.01 and maxq=0.99 nex extremes are hence corrected according to the correction ***
            # *** function for the 1st and 99th percentile.                                                           ***
            ix		<- which.min(abs(cor-series[i]))
            # *** Compute quantile index ***
            series_q[i]	<- minq + (ix-1) * ((maxq-minq)/(npoints-1))
            # *** Compute correction function ***
            cf		<- approx(cor.fun[doy,],n=npoints,method="linear")$y
            # *** Compute corrected value ***
            series_c[i]	<- series[i]-cf[ix]
          }
        }
        
      }
    }
    
    # *** for huss (specific humidity), sfcWind (wind speed), hurs (relative humidity), rsds (global radiation): set negative values to 0 ***
    #if ((var == 'huss') || (var == 'sfcWind') || (var == 'hurs') || (var == 'rsds')) series_c[which(series_c<0)] <- 0
    
    return(list(qm.input.series = series, qm.corrected.series = series_c, quantile.index  = series_q))
  }
}



qm.doywindowgen	<- function(window.width){
  
  doys 		<- c(1:365)
  ndays 		<- length(doys)
  half.width	<- (window.width/2) - 0.5
  
  doy_ix <- array(data=NA,dim=c(ndays,window.width))
  temp.vec <- rep(1:ndays,3)
  
  for (day in doys) { doy_ix[day,] <- temp.vec[((ndays+day)-half.width):((ndays+day)+half.width)] }
  
  return(list(half.width = half.width,window.width = window.width, doy.matrix = doy_ix))
}

qm.doystring <- function(yearbegin,yearend){
  
  doyreg 		<- c(1:365)
  doyleap		<- c(1:59,999,60:365)
  doyscenario     <- integer()
  
  for (countyear in yearbegin:yearend) {
    if (leap_year(countyear)) doyscenario <- c(doyscenario,doyleap) else doyscenario <- c(doyscenario,doyreg)
  }
  return(doyscenario)
}
