"""Reproject and resample Cordex data to target grid."""

# !/usr/bin/env python
# -*- coding: utf-8 -*-

# builtins
import re
import sys
import logging
from joblib import Parallel, delayed
from logging.config import dictConfig

# externals
import numpy as np
import xarray as xr

# locals
from pysegcnn.core.logging import log_conf
from pysegcnn.core.trainer import LogConfig
from climax.core.utils import (get_inventory, reproject_cdo,
                               _parse_cordex_time_span)
from climax.core.cli import preprocess_parser
from climax.core.constants import EUROCORDEX_DOMAIN, CORDEX_PARAMETERS

# module level logger
LOGGER = logging.getLogger(__name__)


def _check_str_arg(arg, pattern):
    # check if input argument matches a defined pattern
    if arg is not None:
        arg_id = re.search(pattern, arg)
        if arg_id is None:
            LOGGER.info('Argument "{}" invalid, form "{}" required.'
                        .format(arg, pattern))
            sys.exit()
        else:
            arg = arg_id[0]

    return arg


if __name__ == '__main__':

    # configure logging
    dictConfig(log_conf())

    # define command line argument parser
    parser = preprocess_parser()

    # parse command line arguments
    args = sys.argv[1:]
    if not args:
        parser.print_help()
        sys.exit()
    else:
        args = parser.parse_args(args)

    # check whether the source directory exists
    if args.source.exists():
        # check whether the target grid file exists
        if not args.grid.exists():
            LOGGER.info('{} does not exist.'.format(args.grid))
            sys.exit()

        # create target directory: check that it is differen from the source
        # directory
        if args.target == args.source:
            LOGGER.info('Source and target paths cannot be equal.')
            sys.exit()

        if not args.target.exists():
            LOGGER.info('mkdir {}'.format(args.target))
            args.target.mkdir(parents=True, exist_ok=True)

        # check ensemble identifier
        ensemble = _check_str_arg(args.ensemble, 'r[0-9]i[0-9]p[0-9]')

        # check regional climate model version
        version = _check_str_arg(args.version, 'v[0-9]')

        # construct file pattern to match from input parameters
        pattern = '(.*).nc$'
        pattern = '_'.join([param if param is not None else '(.*)' for param in
                            [args.variable, EUROCORDEX_DOMAIN, args.gcm,
                             args.scenario, ensemble, args.rcm, version,
                             pattern]])

        # get all the files matching the defined pattern in the source
        # directory
        source = sorted(get_inventory(args.source, pattern, return_df=False))

        # log list of input files
        LogConfig.init_log('Files matching "{}" for variable "{}":'.format(
            pattern, args.variable))
        LOGGER.info(('\n ' + (len(__name__) + 1) * ' ').join(
                    ['{}'.format(file) for file in source]))

        # generate target filenames
        # output path: target/var/scenario
        if (args.variable is not None) and (args.scenario is not None):
            # if variable and scenarios are specified, directly build
            # output path
            output_path = args.target.joinpath(args.variable).joinpath(
                args.scenario)

            # check if output path exists
            if not output_path.exists():
                LOGGER.info('mkdir {}'.format(output_path))
                output_path.mkdir(parents=True, exist_ok=True)

            # list of output files
            target = [output_path.joinpath(file.name) for file in source]
        else:
            # if variable and scenarios are not specified, infer output
            # path from file names
            target = []
            for file in source:
                # parts: dictionary of file name components
                parts = {k: v for k, v in zip(CORDEX_PARAMETERS,
                                              file.stem.split('_'))}

                # construct output path for current file
                output_path = args.target.joinpath(
                    parts['Variable']).joinpath(parts['Scenario'])

                # check if output path exists
                if not output_path.exists():
                    LOGGER.info('mkdir {}'.format(output_path))
                    output_path.mkdir(parents=True, exist_ok=True)

                # output file name
                target.append(output_path.joinpath(file.name))

        # log list of output files
        LogConfig.init_log('Output file names')
        LOGGER.info(('\n ' + (len(__name__) + 1) * ' ').join(
                    ['{}'.format(file) for file in target]))

        # check whether to only print which files would be processed
        if not args.dry_run:
            # run reprojection in parallel
            target = Parallel(n_jobs=-1, verbose=51)(
                delayed(reproject_cdo)(args.grid, src, trg, args.mode,
                                       args.overwrite)
                for src, trg in zip(source, target))

        # check whether to aggregate the netcdf files of a simulation covering
        # differnt time periods into a single file
        if args.aggregate:
            LogConfig.init_log('Aggregating time periods of simulations.')
            # list of unique simulations
            simulations = np.unique([file.stem.rpartition('_')[0] for file in
                                     target])

            # group the list of output files by model simulation
            for sim in simulations:

                # chronologically sorted list of current model simulation
                group = [file for file in target if file.name.startswith(sim)]
                group = sorted(group)

                # create filename for netcdf covering the entire time period of
                # the current simulation
                y_min, _ = _parse_cordex_time_span(group[0])  # first year
                _, y_max = _parse_cordex_time_span(group[-1])  # last year
                filename = '_'.join([sim, '-'.join([y_min, y_max])]) + '.nc'
                filename = group[0].parent.joinpath(filename)

                # log simulation name, time span and files
                LOGGER.info('Aggregating simulation: {}, Time span: {}'
                            .format(sim, '-'.join([y_min, y_max])))
                LOGGER.info(('\n ' + (len(__name__) + 1) * ' ').join(
                    ['{}'.format(file) for file in group]))

                # read multiple netcdf files using xarray and dask
                ds = xr.open_mfdataset(group)

                # save aggregated netcdf file
                LOGGER.info('Saving aggregated NetCDF: {}'.format(filename))
                ds.to_netcdf(filename)

                # remove single netcdf files from disk
                if args.remove:
                    LOGGER.info('Removing individual NetCDF files ...')
                    for file in group:
                        file.unlink()
                        LOGGER.info('rm {}'.format(file))

    else:
        LOGGER.info('{} does not exist.'.format(str(args.source)))
        sys.exit()
